#include <stdio.h>
#include <string.h>

#include "pslib.h"
#include "psAdditionals.h"

#include "pmConfigCameraFromHeader.h"


static void memPrint(const void *ptr)
{
    psMemBlock *mb = ((psMemBlock*)ptr) - 1;
    printf("Memory block %lld (%lld):\n"
	   "\tFile %s, line %d, size %d\n"
	   "\tPosts: %x %x %x\n",
	   mb->id, mb->refCounter, mb->file, mb->lineno, mb->userMemorySize, mb->startblock, mb->endblock,
	   *(void**)((int8_t *)(mb + 1) + mb->userMemorySize));
}

