#ifndef PM_CAMERA_FROM_HEADER_H
#define PM_CAMERA_FROM_HEADER_H

// Does the header match the camera?
bool pmConfigValidateCamera(const psMetadata *camera, const psMetadata *header);

// Work out what camera we have, based on the FITS header and a set of rules specified in the IPP
// configuration; return the camera configuration
psMetadata *pmCameraFromHeader(const psMetadata *header, // The FITS header
			       const psMetadata *ipprc // The IPP configuration
    );

#endif
