#ifndef PM_FPA_GET_VALUE_H
#define PM_FPA_GET_VALUE_H

#include "pslib.h"
#include "pmFPA.h"
#include "psAdditionals.h"

// Look for a particular value for a given cell (referred to by FPA+chip+cell)
psMetadataItem *psCellGetValue(pmCell *cell, // The cell
			       const char *valueName // Name of value
    );

// Type-specific functions provided as a convenience to the user
float psCellGetValueF32(pmCell *cell, // The cell
			const char *valueName // Name of value
    );
int psCellGetValueS32(pmCell *cell, // The cell
		      const char *valueName // Name of value
    );
double psCellGetValueF64(pmCell *cell, // The cell
			 const char *valueName // Name of value
    );
psString psCellGetValueString(pmCell *cell, // The cell
			      const char *valueName // Name of value
    );




#endif
