#ifndef PM_FPA_CONSTRUCT_H
#define PM_FPA_CONSTRUCT_H

#include "pslib.h"
#include "pmFPA.h"

// Read the contents of a FITS file (format specified by the camera configuration) into memory
pmFPA *pmFPAConstruct(const psMetadata *camera, // The camera configuration
		      psDB *db		// Database handle
    );

// Print out the FPA
void pmFPAPrint(pmFPA *fpa		// FPA to print
    );


#endif
