#ifndef PM_CONFIG_H
#define PM_CONFIG_H

#include "pslib.h"

bool pmConfigRead(psMetadata **site, psMetadata **camera, psMetadata **recipe,
                  int *argc, char **argv, const char *recipeName);
bool pmConfigValidateCamera(const psMetadata *camera, const psMetadata *header);
psMetadata *pmConfigCameraFromHeader(const psMetadata *site, const psMetadata *header);
psMetadata *pmConfigRecipeFromCamera(const psMetadata *camera, const char *recipeName);

#endif
